#include <string.h>
#include "elfrun.h"

#ifndef NULL
# define NULL ((void*)0)
#endif

const Elf * elf_detect(const void *p) {
	if (is_elf((const Elf*)p) && !elf_check_executable((const Elf*)p))
		return (const Elf*)p;
	return NULL;
}

const char * elf_get_appname(const Elf *e) {
	const Elf32_Shdr *sh = elf_find_shdr(e, ".appname");
	if (!sh) return "(no name)";
	return (const char*)e + sh->sh_offset;
}

static const void ** elf_get_got_va(const Elf *e) {
	const Elf32_Sym *sym;
	sym = elf_find_symbol(e, "_GLOBAL_OFFSET_TABLE_");
	if (!sym) return NULL;
	return (const void **)sym->st_value;
}

static unsigned short int elf_get_got_size(const Elf *e, const void **gotva) {
	const Elf32_Shdr *sh = elf_find_shdr(e, ".got");
	const void **end;

	if (!sh || !gotva) return 0;

	end = (const void**)((const char*)sh->sh_addr + sh->sh_size);

	return (end < gotva) ? 0 : end - gotva;
}

static void elf_find_writable_va_region(const Elf *e, const void **dbegin, const void **dend) {
	short int i;

	*dbegin = (void*)-1;
	*dend = NULL;
	for (i = 0; i < elf_get_num_phdrs(e); ++i) {
		const Elf32_Phdr *ph = elf_get_phdr(e, i);
		if (ph->p_type != PT_LOAD || !(ph->p_flags & PF_W))
			continue;

		if (ph->p_vaddr < *dbegin)
			*dbegin = ph->p_vaddr;
		if ((const void*)((const char*)ph->p_vaddr + ph->p_memsz) > *dend)
			*dend = (const void*)((const char*)ph->p_vaddr + ph->p_memsz);
	}

	if (*dbegin == (void*)-1)
		*dbegin = NULL;
}

/*static const void * elf_get_builtin_data_start(const Elf *e, const void *minva) {
	const Elf32_Phdr *ph;
	if (!minva) return NULL;
	ph = elf_find_phdr_by_va(e, minva);
	if (!ph) return NULL;
	return (const void*)((const char*)e + (unsigned int)ph->p_offset);
}*/

#define VA2RA(va, ofs)	((void*)((char*)(va) + (ofs)))

static void patch_got(const Elf *e, void **got, unsigned short int size, unsigned long rwofs) {
	const Elf32_Phdr *ph;

	for (; size--; ++got) {
		ph = elf_find_phdr_by_va(e, *got);
		if (!ph) continue;

		*got = VA2RA(*got, (ph->p_flags & PF_W) ? rwofs : (unsigned long)e);
	}
}

static void copy_data(const Elf *e, void *data_start, const void *minva) {
	short int i;
	for (i = 0; i < elf_get_num_phdrs(e); ++i) {
		const Elf32_Phdr *ph = elf_get_phdr(e, i);
		char *dst;
		if (ph->p_type != PT_LOAD || !(ph->p_flags & PF_W)) continue;

		dst = (char*)data_start + (unsigned long)ph->p_vaddr - (unsigned long)minva;
		memcpy(dst, (char*)e + ph->p_offset, ph->p_filesz);
		if (ph->p_memsz > ph->p_filesz)
			memset(dst + ph->p_filesz, 0, ph->p_memsz - ph->p_filesz);
	}
}

int elf_run(const Elf *e) {
	register int ret asm("%d0");
	register const void ** got asm("%a5");
	unsigned short int gots;
	const void *minva, *maxva, **gotva;
	unsigned long rwofs;
	extern char used_ram_end;

	if (!elf_detect(e))
		return -1;

	gotva = elf_get_got_va(e);

	elf_find_writable_va_region(e, &minva, &maxva);

	gots = elf_get_got_size(e, gotva);

	rwofs = (char*)&used_ram_end - (char*)minva;

	// kopira data odmah iza zadnjeg biosovog iskoristenog bajta
	copy_data(e, &used_ram_end, minva);

	got = (const void**)VA2RA(gotva, rwofs);
	patch_got(e, (void**)got, gots, rwofs);

	asm volatile("jsr (%1)"
			: "=d"(ret)
			: "a"(elf_get_entry_ra(e)),
			  "a"(got)
			: "%d0");
	return ret;
}
