#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <poll.h>

#define eprintf(...) fprintf(stderr, __VA_ARGS__)

#define BUF_SIZE	129

static const char DEFAULT_DEVICE[] = "/dev/ttyS0";

static int doreade(int fd, char *buf, int n, int timeout, int report_error) {
	int i, u, t = 0;
	struct pollfd p[1];

	p[0].fd = fd;
	p[0].events = POLLIN;

	for (u = 0; u < n; u += i) {
		p[0].revents = 0;

		if ((i = poll(p, 1, 100)) < 0) {
			if (report_error) eprintf("Poll error!\n");
                        return i;
		}
		if (!i) {
			if (++t > timeout) {
				if (report_error) eprintf("Timeout!\n");
				return 1;
			} else {
				continue;
			}
		}
		t = 0;

		i = read(fd, buf + u, n - u);
		if (i < 0) {
			if (report_error) eprintf("Read failed!\n");
			return i;
		}
	}
	return 0;
}

static int doread(int fd, char *d, int n) {
	return doreade(fd, d, n, 10, 1);
}

static int dowrite(int fd, const char *d, int n) {
	int u, i;
	for (u = 0; u < n; u += i) {
		i = write(fd, d + u, n - u);
		if (i < 0) {
			eprintf("Write error!\n");
			return i;
		}
	}
	return 0;
}

static char *findarg(const char *a, int argn, char *argv[]) {
	int i;
	for (i = 1; i < argn; ++i)
		if (*argv[i] == '-' && !strncmp(a, argv[i] + 1, strlen(a)))
			return argv[i] + strlen(a) + 1;
	return NULL;
}

static char *argfname(int argn, char *argv[]) {
	int i;
	for (i = 1; i < argn; ++i)
		if (*argv[i] != '-')
			return argv[i];
	return NULL;
}

static int open_comm(const char *device) {
	struct termios o;
	int dd;
	char c;

	dd = open(device, O_RDWR | O_NOCTTY | O_NDELAY);

	if (tcgetattr(dd, &o) < 0) goto err;

	cfsetispeed(&o, B9600);
	cfsetospeed(&o, B9600);
	o.c_cflag |= CLOCAL | CREAD;
	o.c_cflag &= ~PARENB;	// no parity
	o.c_cflag &= ~CSTOPB;	// 1 stop bit
	o.c_cflag &= ~CSIZE;
	o.c_cflag |= CS8;	// 8 bit

	o.c_cflag &= ~CRTSCTS;
	o.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);	// raw input
	o.c_iflag &= ~(IXON | IXOFF | IXANY);
	o.c_oflag &= ~OPOST;

	if (tcsetattr(dd, TCSANOW, &o) < 0) goto err;

	// pokupi i ignorira smece iz buffera:
	dowrite(dd, "\n", 1);
	while (!(doreade(dd, &c, 1, 2, 0)));

	return dd;

err:
	close(dd);
	return -1;
}

static int init_comm(int sd) {
	int i;
	char b;

	for (i = 0; i < 5; ++i) {
		dowrite(sd, "Z", 1);
		doread(sd, &b, 1);
		if (b == 'Z') goto ok_kraj;
	}

	return 1;

ok_kraj:;
	dowrite(sd, "\n", 1);
	doread(sd, &b, 1);

	printf("Init ok!\n");

	return 0;
}

int send_file(FILE *fjl, int sd) {
	char buf[BUF_SIZE], chkbuf[BUF_SIZE];
	int sl, ln;
	char *t;
	char check;

	for (ln = 1; fgets(buf, sizeof(buf), fjl); ++ln) {
		if ((t = strchr(buf, '#'))) *t = 0;
		if ((check = (buf[0] == '-')))
			memmove(buf, buf+1, strlen(buf));
		if (sscanf(buf, "sleep %d", &sl) == 1) {
			printf("Sleeping %d seconds...\n", sl);
			sleep(sl);
			continue;
		}
		sl = strlen(buf);
		dowrite(sd, buf, sl);	// TODO neki check
		if (doread(sd, chkbuf, sl)) {
			printf("Read error!\n");
			return 1;
		}
		chkbuf[sl] = 0;
		printf("Line %d: ", ln); fflush(stdout);
		if (!check || !strcmp(buf, chkbuf)) {	// TODO opametiti
			printf("OK\r");
			fflush(stdout);
		} else {
			printf("Err\n");
			printf("\"%s\" <-> \"%s\"\n", buf, chkbuf);
			return 2;
		}
		usleep(10000);		// 10ms sleep
	}

	printf("Done! :)                                      \n");

	return 0;
}

int main(int argn, char *argv[]) {
	const char *fn, *device;
	FILE *fjl;
	int sd;

	// file to send
	fn = argfname(argn, argv);

	// open file
	if (fn) {
		fjl = fopen(fn, "rt");
		if (!fjl) {
			eprintf("Can't open input file!\n");
			return 1;
		}
	} else {
		fjl = stdin;
	}

	// device file
	device = findarg("d", argn, argv);
	if (!device) device = DEFAULT_DEVICE;

	// open
	sd = open_comm(device);
	if (sd < 0) {
		eprintf("Can't open device %s!\n", device);
		fclose(fjl);
		return 1;
	}

	// init CPU comm speed
	if (!findarg("n", argn, argv)) {
		if (init_comm(sd)) {
			eprintf("Can't init communication!\n");
			close(sd);
			fclose(fjl);
			return 1;
		}
	}

	if (send_file(fjl, sd)) {
		eprintf("Error while sending file!\n");
		close(sd);
		fclose(fjl);
		return 1;
	}

	close(sd);
	fclose(fjl);

	return 0;
}
