//! Opis ELF formata
/*!
 * \file elf.h
 * \author Davor Cihlar
 */
#ifndef _ELF_H_
#define _ELF_H_

typedef void *Elf32_Addr;
typedef short int Elf32_Half;
typedef unsigned long Elf32_Off;
typedef long Elf32_Sword;
typedef unsigned long Elf32_Word;

#define ET_NONE		0
#define ET_REL		1
#define ET_EXEC		2
#define ET_DYN		3
#define ET_CORE		4
#define ET_LOPROC	0xFF00
#define ET_HIPROC	0xFFFF

#define EM_NONE		0
#define EM_M32		1
#define EM_SPARC	2
#define EM_386		3
#define EM_68K		4
#define EM_88K		5
#define EM_860		7
#define EM_MIPS		8

#define EI_MAG0		0
#define EI_MAG1		1
#define EI_MAG2		2
#define EI_MAG3		3
#define EI_CLASS	4
#define EI_DATA		5
#define EI_VERSION	6
#define EI_PAD		7
#define EI_NIDENT	16

#define ELFMAG0		0x7F
#define ELFMAG1		'E'
#define ELFMAG2		'L'
#define ELFMAG3		'F'

#define ELFCLASSNONE	0
#define ELFCLASS32	1
#define ELFCLASS64	2

#define ELFDATANONE	0
#define ELFDATA2LSB	1
#define ELFDATA2MSB	2

#define PF_R		0x4
#define PF_W		0x2
#define PF_X		0x1

typedef enum {
	EV_NONE = 0,
	EV_1,
	EV_CURRENT = EV_1
} Elf32_Version;

typedef struct {
        unsigned char	e_ident[EI_NIDENT];	//!< inicijalni podaci
	Elf32_Half	e_type;			//!< tip elf-a
	Elf32_Half	e_machine;		//!< arhitektura
	Elf32_Word	e_version;		//!< verzija elf-a
	Elf32_Addr	e_entry;		//!< virtualna adresa prve instrukcije
	Elf32_Off	e_phoff;		//!< program header offset
	Elf32_Off	e_shoff;		//!< section header offset
	Elf32_Word	e_flags;		//!< prodessor-specific flags
	Elf32_Half	e_ehsize;		//!< velicina elf headera
	Elf32_Half	e_phentsize;		//!< velicina elementa program header-a
	Elf32_Half	e_phnum;		//!< broj elemenata u program headeru
	Elf32_Half	e_shentsize;		//!< velicina section headera
	Elf32_Half	e_shnum;		//!< broj elemenata u section headeru
	Elf32_Half	e_shstrndx;		//!< section header table index of the entry associated with the section name string table
} __attribute__((__packed__)) Elf32_Ehdr;

typedef enum {
	PT_NULL = 0,
	PT_LOAD,
	PT_DYNAMIC,
	PT_INTERP,
	PT_NOTE,
	PT_SHLIB,
	PT_PHDR,
	PT_LOPROC = 0x70000000,
	PT_HIPROC = 0x7FFFFFFF
} Elf32_Phdr_Type;

typedef enum {
	SHT_NULL = 0,
	SHT_PROGBITS,
	SHT_SYMTAB,
	SHT_STRTAB,
	SHT_RELA,
	SHT_HASH,
	SHT_DYNAMIC,
	SHT_NOTE,
	SHT_NOBITS,
	SHT_REL,
	SHT_SHLIB,
	SHT_DYNSYM,
	SHT_LOPROC = 0x70000000,
	SHT_HIPROC = 0x7fffffff
} Elf32_Shdr_Type;

typedef struct {
	Elf32_Word	p_type;			//!< tip
	Elf32_Off	p_offset;		//!< offset od pocetka fajle do segmenta
	Elf32_Addr	p_vaddr;		//!< virtualna adresa segmenta
	Elf32_Addr	p_paddr;		//!< fizikalna adresa segmenta
	Elf32_Word	p_filesz;		//!< duzina validnih podataka u fajli za kopirati u segment
	Elf32_Word	p_memsz;		//!< duzina validnih podataka u segmentu
	Elf32_Word	p_flags;		//!< flagovi
	Elf32_Word	p_align;		//!< string sekcija s nazivima sekcija
} __attribute__((__packed__)) Elf32_Phdr;

typedef struct {
	Elf32_Word	sh_name;
	Elf32_Word	sh_type;
	Elf32_Word	sh_flags;
	Elf32_Addr	sh_addr;
	Elf32_Off	sh_offset;
	Elf32_Word	sh_size;
	Elf32_Word	sh_link;
	Elf32_Word	sh_info;
	Elf32_Word	sh_addralign;
	Elf32_Word	sh_entsize;
} __attribute__((__packed__)) Elf32_Shdr;

typedef struct {
	Elf32_Word	st_name;
	Elf32_Addr	st_value;
	Elf32_Word	st_size;
	unsigned char	st_info;
	unsigned char	st_other;
	Elf32_Half	st_shndx;
} __attribute__((__packed__)) Elf32_Sym;

typedef struct {
	Elf32_Ehdr header;
	char data[1];
} Elf;

typedef enum {
	ELFE_OK = 0,
	ELFE_NOT_ELF,
	ELFE_NOT_32BIT,
	ELFE_NOT_LITTLE_ENDIAN,
	ELFE_NOT_BIG_ENDIAN,
	ELFE_INVALID_VERSION,
	ELFE_NOT_EXE,
	ELFE_NOT_386,
	ELFE_NOT_68K,
	ELFE_NO_PROGRAM_HEADER,
	ELFE_CRIPPLED_HEADER,
} Elf_Error;

char is_elf(const Elf *elf);
short int elf_check_executable(const Elf *elf);
const void * elf_get_entry_ra(const Elf *prog);
short int elf_get_num_phdrs(const Elf *prog);
short int elf_get_num_shdrs(const Elf *prog);
const Elf32_Phdr* elf_get_phdr(const Elf *prog, short int id);
const Elf32_Shdr* elf_get_shdr(const Elf *prog, short int id);
const char * elf_get_string(const Elf *prog, const Elf32_Shdr *sh, short int id);
const Elf32_Shdr * elf_find_shdr(const Elf *prog, const char *name);
const Elf32_Sym * elf_find_symbol(const Elf *prog, const char *name);
const Elf32_Phdr* elf_find_phdr_by_va(const Elf *prog, const void *va);

#endif
